//-----------------------------------------------------------------------------
/*
Programmed by Olivier Sangala
http://olivier.sangala.free.fr
olivier_sangala@hotmail.com
*/
//-----------------------------------------------------------------------------

#include <tigcclib.h>
#include "labyr.h"

//-----------------------------------------------------------------------------

TTile Tiles[MAP_H][MAP_W];
TTCouple TC_LIST[ (MAP_W*(MAP_H-1)) + (MAP_H*(MAP_W-1)) ];

int anti_maze;

//-----------------------------------------------------------------------------

// 'FindWay' variables
TStep StepsL[MAP_H*MAP_W];
int StepCounter;

// Positions
int sx,sy,dx,dy;

//-----------------------------------------------------------------------------

char*Plane0;

//-----------------------------------------------------------------------------

char WallL_Data[8]={0x80,0x80,0x80,0x80,0x00,0x00,0x00,0x00,};
char WallU_Data[8]={0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,};
char CursorData[8]={0x00,0x20,0x70,0x20,0x00,0x00,0x00,0x00,};

//-----------------------------------------------------------------------------

void Initialize(void)
{
	int x,y;
	int counter=0;
	
	for (y=0;y<MAP_H;y++)
	for (x=0;x<MAP_W;x++)
	{
		Tiles[y][x].Owner=NULL;
		Tiles[y][x].WallsFlag=0xFF; // Walls everywhere
		Tiles[y][x].ID=counter++;
	}
}

//-----------------------------------------------------------------------------

TTile* GetOwner(TTile*Tile)
{
	
	if (!Tile->Owner)
	return Tile;
	else
	return GetOwner(Tile->Owner);
	
}
//-----------------------------------------------------------------------------

void TC_Connect(TTCouple*TC)
{
	
	TTile*TR1=GetOwner(TC->Tile1);
	TTile*TR2=GetOwner(TC->Tile2);
	
	// speed up
	if (TR1!=TC->Tile1)
	TC->Tile1->Owner=TR1;
	if (TR2!=TC->Tile2)
	TC->Tile2->Owner=TR2;
	
	// if already connected -> return
	if (TR1->ID==TR2->ID)
	return;
	
	// connect (TR2->TR1)
	TR2->Owner=TR1;
	
	char flag1=(TC->WallType==WO_H?WF_D:WF_R);
	char flag2=(TC->WallType==WO_H?WF_U:WF_L);
	
	TC->Tile1->WallsFlag&=~flag1;
	TC->Tile2->WallsFlag&=~flag2;
	
}
//-----------------------------------------------------------------------------

void MakeMaze(void)
{
	
	DrawStr(0,0,"Please Wait...",A_REPLACE);
	
	int x,y;
	int TC_Count=0;
	
	
	TTCouple*TC=TC_LIST;
	
	// HORIZONTAL
	for (y=0;y<(MAP_H-1);y++)
	for (x=0;x<MAP_W;x++)
	{
		TC->Tile1=&Tiles[y][x];
		TC->Tile2=&Tiles[y+1][x];
		TC->WallType=WO_H;
		
		TC++;
		TC_Count++;
	}
	
	// VERTICAL
	for (y=0;y<MAP_H;y++)
	for (x=0;x<(MAP_W-1);x++)
	{
		TC->Tile1=&Tiles[y][x];
		TC->Tile2=&Tiles[y][x+1];
		TC->WallType=WO_V;
		
		TC++;
		TC_Count++;
	}
	
	// EMPTY QUEUE
	
	for (;;)
	{
		int r=random(TC_Count);
		TC_Connect(&TC_LIST[r]);
		
		TC_LIST[r]=TC_LIST[TC_Count-1];
		TC_Count--;
		
		if (!TC_Count)
		break;
	}
	
}
//-----------------------------------------------------------------------------

void NewMaze(void)
{
	Initialize();
	MakeMaze();
	
	// Positions
	sx=0;
	sy=0;
	dx=(MAP_W-1);
	dy=(MAP_H-1);
	
}
//-----------------------------------------------------------------------------

void FindWay(int x1, int y1, int x2, int y2)
{
	TStep*StepPtr=&StepsL[0];
	
	StepPtr->x=x1;
	StepPtr->y=y1;
	StepPtr->WallTo=-1;
	StepPtr->WallFrom=-1;
	
	StepCounter=0;
	
	for (;;)
	{
		// WAY FOUND ?
		if (StepPtr->x==x2 && StepPtr->y==y2)
		break;
		
		
		// NEXT WALL
		StepPtr->WallTo++;
		
		
		// DIRECTIONS LEFT ?
		if (StepPtr->WallTo<4)
		{
			// CHECK NOT TO GO BACK
			if (StepPtr->WallFrom==-1 || (StepPtr->WallFrom!=StepPtr->WallTo))
			{
				char flag=Tiles[StepPtr->y][StepPtr->x].WallsFlag;
				
				// TRY WAY IF NO WALL
				if ((StepPtr->WallTo==WALL_L && (!(flag&WF_L)))
				||  (StepPtr->WallTo==WALL_R && (!(flag&WF_R)))
				||  (StepPtr->WallTo==WALL_U && (!(flag&WF_U)))
				||  (StepPtr->WallTo==WALL_D && (!(flag&WF_D)))
				)
				{
					TStep*StepPtr2=StepPtr+1;
					*StepPtr2=*StepPtr;
					StepPtr=StepPtr2;
					
					
					// MOVE TO NEXT TILE
					if (StepPtr->WallTo==WALL_L)
					StepPtr->x--;
					else if (StepPtr->WallTo==WALL_R)
					StepPtr->x++;
					else if (StepPtr->WallTo==WALL_U)
					StepPtr->y--;
					else if (StepPtr->WallTo==WALL_D)
					StepPtr->y++;
					
					StepPtr->WallFrom=3-StepPtr->WallTo;
					StepPtr->WallTo=-1;
					
					StepCounter++;
				}
			}
		}
		// ALL DIRECTIONS TRIED.
		else
		{
			StepPtr--;
			StepCounter--;
		}
		
	}
}
//-----------------------------------------------------------------------------


void DrawScreen(void)
{
	
	PortSet(Plane0,239,127);
	ClrScr();
	
	// Tiles
	int x,y;
	for (x=0;x<MAP_W;x++)
	for (y=0;y<MAP_H;y++)
	{
		int xx=(x<<2);
		int yy=(y<<2);
		char flag=Tiles[y][x].WallsFlag;
		
		if (anti_maze)
		flag^=0xFF;
		
		if (flag&WF_L)
		Sprite8(xx,yy,4,WallL_Data,Plane0,SPRT_OR);
		if (flag&WF_U)
		Sprite8(xx,yy,4,WallU_Data,Plane0,SPRT_OR);
	}
	
	// Border Lines
	//DrawLine(0,4*MAP_H,4*MAP_W,4*MAP_H,A_NORMAL);
	//DrawLine(4*MAP_W,0,4*MAP_W,4*MAP_H,A_NORMAL);
	
	// Destination
	Sprite8(dx<<2,dy<<2,4,CursorData,Plane0,SPRT_XOR);
	
	PortRestore();
	memcpy(LCD_MEM,Plane0,LCD_SIZE);
	
}
//-----------------------------------------------------------------------------


// Main Function
void _main(void)
{
	
	Plane0=malloc(LCD_SIZE);
	
	clrscr();
	FontSetSys(0);
	printf("Labyr 1.1 \n");
	printf("Programmed by Olivier Sangala \n");
	printf("http://olivier.sangala.free.fr \n");
	printf("olivier_sangala@hotmail.com \n");
	FontSetSys(1);
	ngetchx();
	
	
	randomize();
	
	anti_maze=0;
	
	NewMaze();
	DrawScreen();
	
	for (;;)
	{
		
		// XOR -> SHOW
		Sprite8(sx<<2,sy<<2,4,CursorData,LCD_MEM,SPRT_XOR);
		// IF SOLVED
		if (sx==dx && sy==dy)
		DrawStr(0,0,"You Win!",A_REVERSE);
		
		int key=ngetchx();
		
		// XOR -> HIDE
		Sprite8(sx<<2,sy<<2,4,CursorData,LCD_MEM,SPRT_XOR);
		// IF SOLVED
		if (sx==dx && sy==dy)
		memcpy(LCD_MEM,Plane0,LCD_SIZE);
		
		
		char flag=Tiles[sy][sx].WallsFlag;
		
		if (key==KEY_LEFT)
		{
		if (sx!=0 && !(flag&WF_L))
		sx--;
		}
		else if (key==KEY_RIGHT)
		{
		if (sx+1!=MAP_W && !(flag&WF_R))
		sx++;
		}
		else if (key==KEY_UP)
		{
		if (sy!=0 && !(flag&WF_U))
		sy--;
		}
		else if (key==KEY_DOWN)
		{
		if (sy+1!=MAP_H && !(flag&WF_D))
		sy++;
		}
		
		else if (key==KEY_F1)
		{
			
			int a;
			FindWay(sx,sy,dx,dy);
			
			// XOR -> SHOW
			for (a=0;a<StepCounter;a++)
			Sprite8(StepsL[a].x<<2,StepsL[a].y<<2,4,CursorData,LCD_MEM,SPRT_XOR);
			
			ngetchx();
			
			// XOR -> HIDE
			for (a=0;a<StepCounter;a++)
			Sprite8(StepsL[a].x<<2,StepsL[a].y<<2,4,CursorData,LCD_MEM,SPRT_XOR);
			
		}
		
		else if (key==KEY_F2)
		{
			NewMaze();
			DrawScreen();
		}
		else if (key==KEY_F3)
		{
			anti_maze^=1;
			DrawScreen();
		}
		else if (key==KEY_F4)
		DrawScreen();
		
		else if (key==KEY_APPS)
		off();
		else if (key==KEY_ESC)
		break;
		
	}
	
	
	
	if (Plane0)
	free(Plane0);
	
}
